/*
 * Copyright (c) 2019 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#![no_std]
#![no_main]
#[link(name="libcommon.rlib")]

extern crate common;

mod uart0;

use common::buffer::Buffer;
use common::clock;
use common::systick;
use common::port;
use common::gpio;
use common::nvic;
use common::i2c;
use common::uart;
use common::usb_serial;
use common::display;
use common::gps;

extern {
    fn enable_interrupts();

    static mut cdc_tx_buf: Buffer;
}

#[no_mangle]
pub unsafe extern "C" fn _start() -> ! {
    clock::configure();
    clock::enable_osc0();
    systick::init();
    port::init();

    // Configure pins for I2C0.
    port::set_af(port::PORTC, 8, 2);
    port::set_af(port::PORTC, 9, 2);

    i2c::configure(i2c::I2C0);

    nvic::disable_irq(8); // I2C0

    // Configure pin for the display's reset line.
    gpio::set_direction(gpio::GPIOB, 1 << 16, gpio::Direction::Output);
    port::set_af(port::PORTB, 16, 1);

    // Configure pins for UART0.
    port::set_af(port::PORTE, 20, 4);
    port::set_af(port::PORTE, 21, 4);

    // Configure pin for the GPS's reset line.
    gpio::set_direction(gpio::GPIOB, 1 << 1, gpio::Direction::Output);
    port::set_af(port::PORTB, 1, 1);

    // Configure upper right push button.
    gpio::set_direction(gpio::GPIOA, 1 << 12, gpio::Direction::Input);
    port::set_af(port::PORTA, 12, 1);
    port::set_pull(port::PORTA, 12, port::Pull::Up);

    enable_interrupts();

    usb_serial::init(0xf055, 0x635d);

    cdc_tx_buf.write(b"\n");
    cdc_tx_buf.flush();

    let mut display = display::Display::new(gpio::GPIOB, 1 << 16, 0x3c);

    display.init();
    display.clear();

    // Hold GPS in reset while configuring its UART.
    gpio::clear(gpio::GPIOB, 1);
    systick::delay_ms(50);
    uart::configure(uart::UART0);
    systick::delay_ms(50);
    gpio::set(gpio::GPIOB, 1);

    nvic::enable_irq(12); // UART0

    let mut gps = gps::Gps::new();

    loop {
        let mut tap = gps::TimeAndPos::new();

        while gps.update(&mut tap) {
        }

        if (gpio::get(gpio::GPIOA) & (1 << 12)) == 0 {
            nvic::system_reset();
        }
    }
}
